@@ -45,15 +45,14 @@
 #include "mongoc-util-private.h"
 #include "mongoc-write-concern-private.h"
 #include "mongoc-uri-private.h"
-
+#ifdef MONGOC_ENABLE_SSL
+#include "mongoc-stream-tls.h"
+#endif
 
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "cluster"
 
 
-#define MIN_WIRE_VERSION 0
-#define MAX_WIRE_VERSION 3
-
 #define CHECK_CLOSED_DURATION_MSEC 1000
 
 #define DB_AND_CMD_FROM_COLLECTION(outstr, name) \
@@ -79,20 +78,6 @@ mongoc_cluster_fetch_stream_pooled (mongoc_cluster_t *cluster,
                                     bool reconnect_ok,
                                     bson_error_t *error);
 
-static const char *
-get_command_name (const bson_t *command)
-{
-   bson_iter_t iter;
-
-   if (!bson_iter_init (&iter, command) ||
-       !bson_iter_next (&iter)) {
-      return NULL;
-   }
-
-   return bson_iter_key (&iter);
-}
-
-
 static void
 _bson_error_message_printf (bson_error_t *error,
                             const char *format,
@@ -120,49 +105,46 @@ _bson_error_message_printf (bson_error_t *error,
 /*
  *--------------------------------------------------------------------------
  *
- * mongoc_cluster_run_command --
+ * mongoc_cluster_run_command_rpc --
  *
- *       Internal function to run a command on a given stream
- *       with read preference PRIMARY.
- *       @error and @reply are optional out-pointers.
+ *       Internal function to run a command on a given stream and
+ *       read the response into @reply_rpc. @rpc and @reply_rpc can be
+ *       the same to reuse storage. @buffer should be initialized before
+ *       passing it in.
  *
  * Returns:
  *       true if successful; otherwise false and @error is set.
  *
  * Side effects:
- *       @reply is set and should ALWAYS be released with bson_destroy().
+ *       On success, @buffer and @reply_rpc are filled out with the reply.
+ *       On failure, @error is filled out. If this was a network error
+ *       and server_id is nonzero, the cluster disconnects from the server.
  *
  *--------------------------------------------------------------------------
  */
 
 bool
-mongoc_cluster_run_command (mongoc_cluster_t            *cluster,
-                            mongoc_stream_t             *stream,
-                            mongoc_query_flags_t         flags,
-                            const char                  *db_name,
-                            const bson_t                *command,
-                            bson_t                      *reply,
-                            bson_error_t                *error)
+mongoc_cluster_run_command_rpc (mongoc_cluster_t    *cluster,
+                                mongoc_stream_t     *stream,
+                                uint32_t             server_id,
+                                const char          *command_name,
+                                mongoc_rpc_t        *rpc,
+                                mongoc_rpc_t        *reply_rpc,
+                                mongoc_buffer_t     *buffer,
+                                bson_error_t        *error)
 {
-   mongoc_buffer_t buffer;
    mongoc_array_t ar;
-   mongoc_rpc_t rpc;
    int32_t msg_len;
-   bson_t reply_local;
-   char ns[MONGOC_NAMESPACE_MAX];
    bool error_set = false;
-   bool reply_local_initialized = false;
    bool ret = false;
+   char db[MONGOC_NAMESPACE_MAX];
 
    ENTRY;
 
    BSON_ASSERT(cluster);
    BSON_ASSERT(stream);
-   BSON_ASSERT(db_name);
-   BSON_ASSERT(command);
 
    _mongoc_array_init (&ar, sizeof (mongoc_iovec_t));
-   _mongoc_buffer_init (&buffer, NULL, 0, NULL, NULL);
 
    if (cluster->client->in_exhaust) {
       bson_set_error(error,
@@ -173,59 +155,114 @@ mongoc_cluster_run_command (mongoc_cluster_t            *cluster,
       GOTO (done);
    }
 
-   bson_snprintf(ns, sizeof ns, "%s.$cmd", db_name);
-
-   rpc.query.msg_len = 0;
-   rpc.query.request_id = ++cluster->request_id;
-   rpc.query.response_to = 0;
-   rpc.query.opcode = MONGOC_OPCODE_QUERY;
-   rpc.query.collection = ns;
-   rpc.query.skip = 0;
-   rpc.query.n_return = -1;
-   rpc.query.fields = NULL;
-   rpc.query.query = bson_get_data (command);
-   rpc.query.flags = flags;
-
-   _mongoc_rpc_gather(&rpc, &ar);
-   _mongoc_rpc_swab_to_le(&rpc);
+   rpc->query.request_id = ++cluster->request_id;
+   _mongoc_rpc_gather (rpc, &ar);
+   _mongoc_rpc_swab_to_le (rpc);
 
    if (!_mongoc_stream_writev_full (stream, (mongoc_iovec_t *)ar.data, ar.len,
                                    cluster->sockettimeoutms, error) ||
-       !_mongoc_buffer_append_from_stream(&buffer, stream, 4,
-                                          cluster->sockettimeoutms, error)) {
+       !_mongoc_buffer_append_from_stream (buffer, stream, 4,
+                                           cluster->sockettimeoutms, error)) {
+
+      mongoc_cluster_disconnect_node (cluster, server_id);
+
       /* add info about the command to writev_full's error message */
+      _mongoc_get_db_name (rpc->query.collection, db);
       _bson_error_message_printf (
          error,
          "Failed to send \"%s\" command with database \"%s\": %s",
-         get_command_name (command),
-         db_name,
-         error->message);
+         command_name, db, error->message);
 
       error_set = true;
       GOTO (done);
    }
 
-   BSON_ASSERT(buffer.len == 4);
+   BSON_ASSERT (buffer->len == 4);
 
-   memcpy(&msg_len, buffer.data, 4);
+   memcpy (&msg_len, buffer->data, 4);
    msg_len = BSON_UINT32_FROM_LE(msg_len);
-   if ((msg_len < 16) || (msg_len > (1024 * 1024 * 16))) {
+   if ((msg_len < 16) || (msg_len > MONGOC_DEFAULT_MAX_MSG_SIZE)) {
       GOTO (done);
    }
 
-   if (!_mongoc_buffer_append_from_stream(&buffer, stream, msg_len - 4,
-                                          cluster->sockettimeoutms, error)) {
+   if (!_mongoc_buffer_append_from_stream (buffer, stream, (size_t) msg_len - 4,
+                                           cluster->sockettimeoutms, error)) {
       error_set = true;
       GOTO (done);
    }
 
-   if (!_mongoc_rpc_scatter(&rpc, buffer.data, buffer.len)) {
+   if (!_mongoc_rpc_scatter (reply_rpc, buffer->data, buffer->len)) {
       GOTO (done);
    }
 
-   _mongoc_rpc_swab_from_le(&rpc);
+   _mongoc_rpc_swab_from_le (reply_rpc);
 
-   if (rpc.header.opcode != MONGOC_OPCODE_REPLY) {
+   if (reply_rpc->header.opcode != MONGOC_OPCODE_REPLY) {
+      GOTO (done);
+   }
+
+   ret = true;
+
+done:
+   _mongoc_array_destroy (&ar);
+
+   if (!ret && !error_set) {
+      /* generic error */
+      bson_set_error (error,
+                      MONGOC_ERROR_PROTOCOL,
+                      MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
+                      "Invalid reply from server.");
+   }
+
+   RETURN (ret);
+}
+
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_cluster_run_command --
+ *
+ *       Internal function to run a command on a given stream.
+ *       @error and @reply are optional out-pointers.
+ *
+ * Returns:
+ *       true if successful; otherwise false and @error is set.
+ *
+ * Side effects:
+ *       @reply is set and should ALWAYS be released with bson_destroy().
+ *       On failure, @error is filled out. If this was a network error
+ *       and server_id is nonzero, the cluster disconnects from the server.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+bool
+mongoc_cluster_run_command (mongoc_cluster_t    *cluster,
+                            mongoc_stream_t     *stream,
+                            uint32_t             server_id,
+                            mongoc_query_flags_t flags,
+                            const char          *db_name,
+                            const bson_t        *command,
+                            bson_t              *reply,
+                            bson_error_t        *error)
+{
+   char ns[MONGOC_NAMESPACE_MAX];
+   mongoc_rpc_t rpc;
+   bson_t reply_local;
+   bool ret = false;
+   bool reply_local_initialized = false;
+   mongoc_buffer_t buffer;
+
+   _mongoc_buffer_init (&buffer, NULL, 0, NULL, NULL);
+
+   bson_snprintf (ns, sizeof ns, "%s.$cmd", db_name);
+
+   _mongoc_rpc_prep_command (&rpc, ns, command, flags);
+
+   /* we can reuse the query rpc for the reply */
+   if (!mongoc_cluster_run_command_rpc (cluster, stream, server_id,
+                                        _mongoc_get_command_name (command),
+                                        &rpc, &rpc, &buffer, error)) {
       GOTO (done);
    }
 
@@ -235,14 +272,12 @@ mongoc_cluster_run_command (mongoc_cluster_t            *cluster,
                       MONGOC_ERROR_BSON,
                       MONGOC_ERROR_BSON_INVALID,
                       "Failed to decode reply BSON document.");
-      error_set = true;
       GOTO (done);
    }
 
    reply_local_initialized = true;
 
    if (_mongoc_rpc_parse_command_error (&rpc, error)) {
-      error_set = true;
       GOTO (done);
    }
 
@@ -256,16 +291,7 @@ done:
       bson_init (reply);
    }
 
-   _mongoc_buffer_destroy(&buffer);
-   _mongoc_array_destroy(&ar);
-
-   if (!ret && !error_set) {
-      /* generic error */
-      bson_set_error (error,
-                      MONGOC_ERROR_PROTOCOL,
-                      MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
-                      "Invalid reply from server.");
-   }
+   _mongoc_buffer_destroy (&buffer);
 
    RETURN (ret);
 }
@@ -305,7 +331,7 @@ _mongoc_stream_run_ismaster (mongoc_cluster_t *cluster,
    bson_init (&command);
    bson_append_int32 (&command, "ismaster", 8, 1);
 
-   ret = mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+   ret = mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                      "admin", &command, reply, error);
 
    bson_destroy (&command);
@@ -499,7 +525,7 @@ _mongoc_cluster_auth_node_cr (mongoc_cluster_t      *cluster,
     */
    bson_init (&command);
    bson_append_int32 (&command, "getnonce", 8, 1);
-   if (!mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+   if (!mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                     auth_source, &command, &reply, error)) {
       bson_destroy (&command);
       bson_destroy (&reply);
@@ -534,7 +560,7 @@ _mongoc_cluster_auth_node_cr (mongoc_cluster_t      *cluster,
     * Execute the authenticate command. mongoc_cluster_run_command
     * checks for {ok: 1} in the response.
     */
-   ret = mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+   ret = mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                      auth_source, &command, &reply, error);
    if (!ret) {
       /* error->message is already set */
@@ -724,7 +750,7 @@ _mongoc_cluster_auth_node_sasl (mongoc_cluster_t *cluster,
                    mongoc_uri_get_username (cluster->uri),
                    sasl.step);
 
-      if (!mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+      if (!mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                        "$external", &cmd, &reply, error)) {
          bson_destroy (&cmd);
          bson_destroy (&reply);
@@ -847,7 +873,7 @@ _mongoc_cluster_auth_node_plain (mongoc_cluster_t      *cluster,
    bson_append_utf8 (&b, "payload", 7, (const char *)buf, buflen);
    BSON_APPEND_INT32 (&b, "autoAuthorize", 1);
 
-   ret = mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+   ret = mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                      "$external", &b, &reply, error);
 
    if (!ret) {
@@ -901,7 +927,7 @@ _mongoc_cluster_auth_node_x509 (mongoc_cluster_t      *cluster,
    BSON_APPEND_UTF8 (&cmd, "mechanism", "MONGODB-X509");
    BSON_APPEND_UTF8 (&cmd, "user", username);
 
-   ret = mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+   ret = mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                      "$external", &cmd, &reply, error);
 
    if (!ret) {
@@ -971,7 +997,7 @@ _mongoc_cluster_auth_node_scram (mongoc_cluster_t      *cluster,
                    mongoc_uri_get_username (cluster->uri),
                    scram.step);
 
-      if (!mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
+      if (!mongoc_cluster_run_command (cluster, stream, 0, MONGOC_QUERY_SLAVE_OK,
                                        auth_source, &cmd, &reply, error)) {
          bson_destroy (&cmd);
          bson_destroy (&reply);
@@ -1076,7 +1102,7 @@ _mongoc_cluster_auth_node (mongoc_cluster_t *cluster,
 
    /* Use cached max_wire_version, not value from sd */
    if (!mechanism) {
-      if (max_wire_version < 3) {
+      if (max_wire_version < WIRE_VERSION_SCRAM_DEFAULT) {
          mechanism = "MONGODB-CR";
       } else {
          mechanism = "SCRAM-SHA-1";
@@ -1417,10 +1443,12 @@ mongoc_cluster_fetch_stream_single (mongoc_cluster_t *cluster,
    mongoc_topology_t *topology;
    mongoc_stream_t *stream;
    mongoc_topology_scanner_node_t *scanner_node;
+   int64_t timeout_ms;
    int64_t expire_at;
    bson_t reply;
 
    topology = cluster->client->topology;
+   timeout_ms = topology->connect_timeout_msec;
 
    scanner_node = mongoc_topology_scanner_get_node (topology->scanner, sd->id);
    BSON_ASSERT (scanner_node && !scanner_node->retired);
@@ -1437,7 +1465,8 @@ mongoc_cluster_fetch_stream_single (mongoc_cluster_t *cluster,
       }
       stream = scanner_node->stream;
 
-      expire_at = bson_get_monotonic_time() + topology->connect_timeout_msec * 1000;
+      expire_at = bson_get_monotonic_time() + timeout_ms * 1000;
+
       if (!mongoc_stream_wait (stream, expire_at)) {
          bson_set_error (error,
                          MONGOC_ERROR_STREAM,
@@ -1449,6 +1478,28 @@ mongoc_cluster_fetch_stream_single (mongoc_cluster_t *cluster,
          return NULL;
       }
 
+#ifdef MONGOC_ENABLE_SSL
+      if (scanner_node->ts->ssl_opts) {
+         if (!mongoc_stream_tls_do_handshake (stream, (int32_t) timeout_ms)) {
+            bson_set_error (error,
+                            MONGOC_ERROR_STREAM,
+                            MONGOC_ERROR_STREAM_SOCKET,
+                            "TLS handshake failed.");
+            mongoc_topology_scanner_node_disconnect (scanner_node, true);
+            return NULL;
+         }
+
+         if (!mongoc_stream_tls_check_cert (stream, scanner_node->host.host)) {
+            bson_set_error (error,
+                            MONGOC_ERROR_STREAM,
+                            MONGOC_ERROR_STREAM_SOCKET,
+                            "Failed to verify peer certificate.");
+            mongoc_topology_scanner_node_disconnect (scanner_node, true);
+            return NULL;
+         }
+      }
+#endif
+
       if (!_mongoc_stream_run_ismaster (cluster, stream, &reply, error)) {
          mongoc_topology_scanner_node_disconnect (scanner_node, true);
          return NULL;
@@ -2020,6 +2071,8 @@ _mongoc_cluster_check_interval (mongoc_cluster_t *cluster,
    if (scanner_node->last_used + (1000 * CHECK_CLOSED_DURATION_MSEC) < now) {
       if (mongoc_stream_check_closed (stream)) {
          mongoc_cluster_disconnect_node (cluster, server_id);
+         bson_set_error (error, MONGOC_ERROR_STREAM, MONGOC_ERROR_STREAM_SOCKET,
+                         "Stream is closed");
          return false;
       }
    }
@@ -2030,9 +2083,9 @@ _mongoc_cluster_check_interval (mongoc_cluster_t *cluster,
       BSON_APPEND_INT32 (&command, "ismaster", 1);
 
       before_ismaster = now;
-
-      r = mongoc_cluster_run_command (cluster, stream, MONGOC_QUERY_SLAVE_OK,
-                                      "admin", &command, &reply, error);
+      r = mongoc_cluster_run_command (cluster, stream, server_id,
+                                      MONGOC_QUERY_SLAVE_OK, "admin", &command,
+                                      &reply, error);
 
       now = bson_get_monotonic_time ();
 
@@ -2143,7 +2196,7 @@ mongoc_cluster_sendv_to_server (mongoc_cluster_t              *cluster,
       need_gle = _mongoc_rpc_needs_gle(&rpcs[i], write_concern);
       _mongoc_rpc_gather (&rpcs[i], &cluster->iov);
 
-      max_msg_size = mongoc_stream_max_msg_size (server_stream);
+      max_msg_size = mongoc_server_stream_max_msg_size (server_stream);
 
       if (rpcs[i].header.msg_len > max_msg_size) {
          bson_set_error(error,
@@ -2279,7 +2332,7 @@ mongoc_cluster_try_recv (mongoc_cluster_t       *cluster,
     */
    memcpy (&msg_len, &buffer->data[buffer->off + pos], 4);
    msg_len = BSON_UINT32_FROM_LE (msg_len);
-   max_msg_size = mongoc_stream_max_msg_size (server_stream);
+   max_msg_size = mongoc_server_stream_max_msg_size (server_stream);
    if ((msg_len < 16) || (msg_len > max_msg_size)) {
       bson_set_error (error,
                       MONGOC_ERROR_PROTOCOL,
