@@ -30,6 +30,10 @@ BSON_BEGIN_DECLS
 bool _mongoc_host_list_from_string (mongoc_host_list_t *host_list,
                                     const char         *host_and_port);
 
+bool _mongoc_host_list_equal (const mongoc_host_list_t *host_a,
+                              const mongoc_host_list_t *host_b);
+
+void _mongoc_host_list_destroy_all (mongoc_host_list_t *host);
 
 BSON_END_DECLS
 
